/*
Copyright © 2020 NAME HERE <EMAIL ADDRESS>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package pkg

import (
	"context"
	"os"
	"sync"

	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/s3/s3manager"
	"github.com/go-logr/logr"
	"github.com/go-logr/zapr"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
	"go.uber.org/zap"
)

var cfgFile string

const (
	// The name of our config file, without the file extension because viper supports many different config file languages.
	defaultConfigFilename = ".config"
	CONNECTIONSTRING      = "mongodb://localhost:27017/rocketchat"
	DB                    = "rocketchat"

	// The environment variable prefix of all environment variables bound to our command line flags.
	// For example, --number is bound to STING_NUMBER.
//	envPrefix = "ROCKETCHAT"

)

var home = os.Getenv("HOME")

type Config struct {
	RocketChatURL string   `json:"rcURL"`
	Collections   []string `json:"collections"`
	S3            *S3      `json:"s3"`
	Mongo         *Mongo   `json:"mongo"`
	/*
		MongoURI      string `json:"hostname"`
		   	Database      string
		   	Bucket        string
		   	S3URL         string
		   	Secret        string
		   	Key           string
	*/
}

type S3 struct {
	Bucket string `json:"bucket"`
	*aws.Config
}

type Mongo struct {
	URI      string `json:"uri"`
	Database string `json:"database"`
	*options.ClientOptions
}

type Client struct {
	log        logr.Logger
	mongo      *mongo.Database
	uploader   *s3manager.Uploader
	rcURL      string
	rcUID      string
	collection string
	bucket     string
}

func NewClient(conf *Config) (*Client, error) {
	var mongoError error
	var mongoOnce sync.Once

	client := &Client{}
	mongoclient := &mongo.Client{}

	sess, err := session.NewSession(conf.S3.Config)
	if err != nil {
		return client, err
	}

	client.uploader = s3manager.NewUploader(sess)

	mongoOnce.Do(func() {
		clientOptions := &options.ClientOptions{}

		if len(conf.Mongo.URI) > 0 {
			clientOptions = options.Client().ApplyURI(conf.Mongo.URI)

		} else {
			clientOptions = conf.Mongo.ClientOptions
		}

		// Connect to MongoDB
		client, err := mongo.Connect(context.TODO(), clientOptions)
		if err != nil {
			mongoError = err
		}
		// Check the connection
		err = client.Ping(context.TODO(), nil)
		if err != nil {
			mongoError = err
		}
		mongoclient = client
	})

	if mongoError != nil {
		return client, mongoError
	}

	client.mongo = mongoclient.Database(conf.Mongo.Database)

	log := zapr.NewLogger(zap.NewExample())
	log = log.WithName("RCS3")

	client.log = log

	ctx := context.TODO()

	client.rcUID = GetUniqueID(ctx, client.mongo)
	client.rcURL = conf.RocketChatURL
	client.bucket = conf.S3.Bucket

	return client, nil
}
