package pkg

import (
	"bytes"
	"context"
	"fmt"
	"io"
	"log"

	"github.com/pkg/errors"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/gridfs"
	"go.mongodb.org/mongo-driver/mongo/options"
)

type Store interface {
	ItemsMap()
	Put(u *Upload)
	Delete(u *Upload)
}

type S3Store struct {
	c *Client
}

type FileSystem struct {
	c *Client
}

type GridFS struct {
	c *Client
}

type Migrator struct {
	Source      Store
	Destination Store
}

func ReadFiles(ctx context.Context, c *Client, coll string) (<-chan Upload, <-chan error, error) {
	collection := c.mongo.Collection(coll)

	var store string
	switch coll {
	case "rocketchat_uploads":
		store = "GridFS:Uploads"
	case "rocketchat_avatars":
		store = "GridFS:Avatars"
	default:
		// TODO improve error
		return nil, nil, errors.New("unknown collection")

	}

	cur, err := collection.Find(ctx, bson.M{"store": store, "complete": true})
	if err != nil {
		return nil, nil, err
	}

	out := make(chan Upload)
	errc := make(chan error, 1)

	total, err := collection.CountDocuments(ctx, bson.M{"store": store, "complete": true})
	if err != nil {
		return nil, nil, err
	}

	c.log.Info("count", "files to migrate", total)

	go func() {
		defer cur.Close(ctx)
		defer close(out)
		defer close(errc)

		if total == 0 {
			c.log.Info("no files to migrate")
			ctx.Done()
			return
		}

		for cur.Next(ctx) {
			result := Upload{}

			err := cur.Decode(&result)
			if err != nil {
				errc <- err
				return
			}

			c.log.WithValues("task", "reading").WithValues("collection", coll).WithValues("file", result.ID).WithValues("size", result.Size).Info("starting")

			// TODO
			result.MutateToS3(c.rcURL, c.rcUID)

			// Send the data to the output channel but return early
			// if the context has been cancelled.
			select {
			case out <- result:
			case <-ctx.Done():
				return
			}
		}
	}()

	return out, errc, nil
}

func DownloadToStream(u *Upload, coll *mongo.Collection) (io.Reader, int64, error) {
	buf := new(bytes.Buffer)

	bucket, err := gridfs.NewBucket(
		coll.Database(),
		options.GridFSBucket().SetName(coll.Name()),
	)
	if err != nil {
		return buf, 0, err
	}

	dStream, err := bucket.DownloadToStreamByName(u.ID, buf)
	if err != nil {
		return buf, dStream, err
	}

	return buf, dStream, nil
}

func MoveFiles(ctx context.Context, c *Client, u *Upload, coll string) error {
	collection := c.mongo.Collection(coll)

	buf, _, err := DownloadToStream(u, collection)
	if err != nil {
		return err
	}

	u.MutateToS3(c.rcURL, c.rcUID)

	uInputs := NewS3UploadInput(u, c.bucket)
	uInputs.Body = buf

	_, err = c.uploader.Upload(uInputs)

	if err != nil {
		return err
	}

	return nil
}

func DeleteFiles(ctx context.Context, c *Client, u *Upload, coll string) error {
	collection := c.mongo.Collection(coll)

	bucket, err := gridfs.NewBucket(
		c.mongo,
		options.GridFSBucket().SetName(collection.Name()),
	)
	if err != nil {
		return err
	}

	if err := bucket.Delete(u.ID); err != nil {
		// If file is already delete we can ignore
		if err != gridfs.ErrFileNotFound {
			return err
		}
	}

	return nil

}

func UpdateDB(ctx context.Context, c *Client, u *Upload, coll string) error {
	collection := c.mongo.Collection(coll)

	filter := bson.M{"_id": u.ID}
	update := bson.M{
		"$set": u,
	}

	_, err := collection.UpdateOne(ctx, filter, update)

	if err != nil {
		return err
	}

	return nil
}

func GetUniqueID(ctx context.Context, db *mongo.Database) string {
	var uid bson.M
	collection := db.Collection("rocketchat_settings")

	if err := collection.FindOne(ctx, bson.M{"_id": "uniqueID"}).Decode(&uid); err != nil {
		log.Fatal(err)
	}

	return fmt.Sprintf("%v", uid["value"])
}
