package pkg

import (
	"context"
	"fmt"
	"sync"

	"github.com/pkg/errors"
)

type Stage func(ctx context.Context, c *Client, in <-chan Upload, coll string) (<-chan Upload, <-chan error, error)

type TaskRunner func(ctx context.Context, c *Client, u *Upload, coll string) error

type Task interface {
	Name() string
	Runner() TaskRunner
}
type task struct {
	runner TaskRunner
	name   string
}

func (t *task) Name() string {
	return t.name
}

func (t *task) Runner() TaskRunner {
	return t.runner
}

func NewTasks(cmd string) map[int]Task {
	tasks := make(map[int]Task)

	updateDBTask := &task{
		name:   "updateDB",
		runner: UpdateDB,
	}

	moveFilesTask := &task{
		name:   "move",
		runner: MoveFiles,
	}

	deleteFilesTask := &task{
		name:   "delete",
		runner: DeleteFiles,
	}

	switch cmd {
	case "move":
		tasks[1] = moveFilesTask
	case "updateDB":
		tasks[1] = updateDBTask
	case "delete":
		tasks[1] = deleteFilesTask
	case "run":
		tasks[1] = moveFilesTask
		tasks[2] = updateDBTask
		tasks[3] = deleteFilesTask
	}
	fmt.Println(tasks)
	return tasks
}

func NewStage(task Task) Stage {
	return func(ctx context.Context, c *Client, in <-chan Upload, coll string) (<-chan Upload, <-chan error, error) {
		out := make(chan Upload)
		errc := make(chan error, 1)
		go func() {
			defer close(out)
			defer close(errc)
			//	c.log.

			for u := range in {
				c.log.WithValues("task", task.Name()).WithValues("collection", coll).WithValues("file", u.ID).WithValues("size", u.Size).Info("starting")

				err := task.Runner()(ctx, c, &u, coll)
				if err != nil {
					// Handle an error that occurs during the goroutine.
					errc <- err
					return
				}
				// Send the data to the output channel but return early
				// if the context has been cancelled.
				select {
				case out <- u:
				case <-ctx.Done():
					return
				}
			}
		}()

		return out, errc, nil
	}

}

func Pipeline(ctx context.Context, c *Client, coll string, tasks map[int]Task) error {
	var errcList []<-chan error

	uploads := make(map[int](<-chan Upload) /* , len(tasks)+1 */)

	u, errc, err := ReadFiles(ctx, c, coll)
	if err != nil {
		return err
	}

	errcList = append(errcList, errc)

	uploads[0] = u

	for i, t := range tasks {
		u, errc, err := NewStage(t)(ctx, c, uploads[i-1], coll)
		if err != nil {
			return err
		}

		uploads[i] = u
		errcList = append(errcList, errc)

	}
	// Sink pipeline stage.
	errc, err = sink(ctx, uploads[len(tasks)])
	if err != nil {
		return err
	}

	errcList = append(errcList, errc)

	return WaitForPipeline(errcList...)
}

func sink(ctx context.Context, in <-chan Upload) (
	<-chan error, error) {
	errc := make(chan error, 1)
	go func() {
		defer close(errc)
		for n := range in {
			if &n == nil {
				// Handle an error that occurs during the goroutine.
				errc <- errors.Errorf("number %v is too large", n)
				return
			}
			//		n.MutateToS3()
			//	fmt.Printf("sink: %v\n", n.URL)
			//	fmt.Print(n)
		}
	}()
	return errc, nil
}

// MergeErrors merges multiple channels of errors.
// Based on https://blog.golang.org/pipelines.
func MergeErrors(cs ...<-chan error) <-chan error {
	var wg sync.WaitGroup
	// We must ensure that the output channel has the capacity to hold as many errors
	// as there are error channels. This will ensure that it never blocks, even
	// if WaitForPipeline returns early.
	out := make(chan error, len(cs))

	// Start an output goroutine for each input channel in cs.  output
	// copies values from c to out until c is closed, then calls wg.Done.
	output := func(c <-chan error) {
		for n := range c {
			out <- n
		}
		wg.Done()
	}
	wg.Add(len(cs))
	for _, c := range cs {
		go output(c)
	}

	// Start a goroutine to close out once all the output goroutines are
	// done.  This must start after the wg.Add call.
	go func() {
		wg.Wait()
		close(out)
	}()

	return out
}

// WaitForPipeline waits for results from all error channels.
// It returns early on the first error.
func WaitForPipeline(errs ...<-chan error) error {
	errc := MergeErrors(errs...)

	for err := range errc {
		if err != nil {
			return err
		}
	}
	return nil
}
